#ifndef SEGMENT_H_INCLUDED_
#define SEGMENT_H_INCLUDED_

#include<vector>
#include<set>
#include<iostream>

#include "rose.h"
#include "boost/graph/properties.hpp" // needed for color

#include "mapped_variable.h" // needed for Conflict

#include "time.h" // needed for Time
#include "port_call_path.h"
#include "../internal_representation/thread.h"

namespace risc {

namespace sg {

class Segment {

public:

  /**
   * \brief   TLM2 attribute
   *          marks if a segment is in a b_transport function
   */
  bool is_b_transport_seg_;
  
  /**
   * \brief   TLM2 attribute
   *          the function that this segment belongs to
   */
  SgFunctionDeclaration* b_transport_func_;
  
  /**
   * \brief   TLM2 attribute
   *          the b_transport call that this segment contains
   */
  SgFunctionCallExp* b_transport_;
  
  /**
   * \brief   PSG_NODE attribute
   *          marks if a node is fake, not a true segment node and will 
   *          be merged
   */
  bool is_fake_segment_;


  /**
   * \brief   PSG_NODE attribute
   *          marks if a node is port call for sc_fifo.
   *          if it is, then during instrumentation, we put the instrumented 
   *          port id into the port call instead of in front of it
   */
  bool is_sc_fifo_port_call_;

  /**
   * \brief   PSG_NODE attribute
   *          marks if a node is for a simulation process
   *          currently, it applies only to the first psg node, because
   *          ZC thinks it is sufficient
   */
  bool is_simulation_process_;


  /**
   * \brief   PSG_NODE attribute
   *          each segment node in a psg has a unique contiuous id
   *          it is used for instrumentation, and also calculating the
   *          wait_offset_
   *          <psg_id_, segment_node_id_> pair matches a segment id in
   *          the complete segment graph
   */
  int segment_node_id_;
  
  /**
   * \brief   PSG_NODE attribute
   *          the qualified name of function that this node is entrance to.
   */
  std::string function_name_entrance_;

  /**
   * \brief   PSG_NODE attribute
   *          the qualified name of function that this node is exit of.
   */
  std::string function_name_exit_;

  /**
   * \brief   PSG_NODE attribute
   *          the qualified name of module that this node belongs to
   *          it is not empty only if the function of the node is a 
   *          simulation process.
   *          currently, it applies only to the first psg node, because
   *          ZC thinks it is sufficient
   */
  std::string module_name_;
  
  /**
   * \brief   PSG_NODE attribute
   *          if the node is partial function call node, this attribute
   *          stores the qualified name of the non-defining function call
   */
  std::string non_defining_function_call_name_;
  
  /**
   * \brief   PSG_NODE attribute
   *          if the node is partial function call node, and the
   *          non-defining function call in the node is a port call,
   *          this attributes stores the qualified name of the port.
   *          when psg is loaded, we can get the definition of the port,
   *          and reconstruct the pcp for the variables that are used to
   *          replace the node
   */
  std::string pcp_port_name_;

  /**
   * \brief   PSG_NODE attribute
   *          reconstructed when a partial function call node is loaded back, 
   *          using the stored pcp_port_name_.
   *          when used, should be converted to sgVariableDefinition
   */
  SgNode* pcp_port_;

  //psg cloning ZC step 1
  /**
   * \brief   PSG_NODE attribute
   *          If a function is a channel method (implemented in sc_channel)
   *          Then is_entrance_to_channel_method_ is set to true for the
   *          first segment node of the function's psg
   */
  bool is_entrance_to_channel_method_;

  //psg cloning ZC step 5
  /**
   * \brief   PSG_NODE attribute
   *          if this node is a partial function call node, and the
   *          non defining function call is a port call,
   *          port_call_func_exp is the func call expression.
   *          during instrumentation, we put set_upcoming_seg_id in front
   */
  SgFunctionCallExp* port_call_func_exp_;

  /**
   * \brief   PSG_NODE attribute
   *          the variable name of the channel segment array. 
   *          After instrumentation, it looks like this:
   *          set_upcoming_seg_id(risc_tmp_"unique_port_call_name_"+wait_offset_)
   *          extern int risc_tmp_"unique_port_call_name_"[]
   *          stored in psg file,
   *          loaded back and reconstructed in partial func call nodes
   *          this makes the instrumented names in both port call file
   *          and top file same
   */
  std::string unique_port_call_name_;
  //end

  //psg cloning ZC step 5
  /**
   * \brief   PSG_NODE attribute
   *          After integration, we use this attribute to identify in which 
   *          channel segment id array will the nodes's id be instrumented
   */
  std::string channel_seg_ids_array_name_;
  //end


  /**
   * \brief   PSG_NODE attribute
   *          id of the psg that this node belongs to          
   */
  int psg_id_;
 
  /**
   * \brief   PSG_NODE attribute
   *          id of node, continuous
   */
  int node_id_;

  /**
   * \brief   PSG_NODE attribute
   *          if a node is removed in the psg file by the provider,
   *          when loaded back, we generate an empty node to represent
   *          the redacted node.
   *          when is_redacted_ is true, the node will not be removed 
   *          from graph_ in remove_dangling_segments().
   *          initially false.
   */
  bool is_redacted_;

  /**
   * \brief   PSG_NODE attribute
   *          a static counter that counts the current id of current node
   */
  static int node_counter;
  
  enum ERead {NonRead = 0, Read = 1};
  enum EWrite {NonWrite = 0, Write = 1};

  //typedef std::vector<const SgStatement*> Statements;
  typedef std::list<NodeWithPath> Expressions;
  boost::default_color_type color;

  enum Status {Untouched = 0, Touched = 1};
  Status color_for_communication_graph_;
  Status helper_for_graph_algorithms_;

  Expressions expressions_;

  bool debugging_;

  /**
   * \brief   if this segemnt should be merged into others
   */
  bool inlined_segment_;

  /**
   * \brief   if this segemnt belongs to a channel function
   */
  bool is_channel_segment_;

  /**
   * \brief   used in instrumentation
   */
  int wait_offset_;

  /**
   * \brief   the conflicts in this segment
   */
  std::set<Conflict> conflicts_;

  /**
   * \fn       Segment(SgNode *boundary_node,
          bool inlined_segment,
          int psg_id = 0,
          bool is_fake_segment_ = false,
          bool is_simulation_process = false)
   * \brief    Constructor for segment.
   *           
   */
  Segment(SgNode *boundary_node,
          bool inlined_segment,
          int psg_id = 0,
          bool is_fake_segment_ = false,
          bool is_simulation_process = false);

  Segment(const Segment &other);

  /**
   * \brief   To identify segments which belong to segments
   */
  PortCallPath pcp_;

  /**
   * \fn       analyze_segment
   * \brief    This function determines which variable are read and written
   *           in this segment
   */
  void analyze_segment();

  /**
   * \fn       analyze_expression(SgNode *expression, ERead read, EWrite write, PortCallPath pcp)
   * \brief    This function analyzes recursively all read and write accesses
   *           in this expression
   */
  void analyze_expression(SgNode *expression, ERead read, EWrite write, PortCallPath pcp);

  friend std::ostream& operator<< (std::ostream &out, const Segment &segment)
  {
    return out;
  }

  /**
   * \fn       add_expression(NodeWithPath nwp)
   * \brief    This function add the given node to the vector of expressions
   */
  void add_expression(NodeWithPath nwp);

  /**
   * \fn       operator==(const Segment &other)
   * \brief    If two segments are equal the return value is true
   */
  bool operator==(const Segment &other);

  /*
   * \brief Merges the other segment in this segment.
   *        No statement will be doubled
   */
  void merge(const Segment &other);

  /**
   * \brief This is the unique id of the segment
   */
  int id_;
  
  /**
   * \brief This is the associated boundary of the segment
   */
  SgNode * boundary_node_;

  /**
   * \brief If known, this is the originating thread of the segment.
   *        If not known, this is NULL.
   *        If multiple threads reach here, this is -1.
   *        TODO: extend for all SystemC processes (RD, 07/11/18)
   */
  Thread *originating_thread_; // (RD, 07/10/18)

  /**
   * \brief This static counter will be used to generate id's for the
   *        individual segments.
   */
  static int id_counter;

  /**
   * \brief This color is used for some boost algorithms
   */
  boost::default_color_type color_;

  /**
   * \brief Set of all SgVariableSymbol which are read in this segment
   */
  std::set<SymbolWithPath> read_variables_;

  /**
   * \brief Set of all SgVariableSymbol which are written in this segment
   */
  std::set<SymbolWithPath> write_variables_;

  /**
   * \brief List of all gathered events this segment is waiting for
   *        TODO: this should be computed separately, not as side-effect of
   *        risc::sg::EventConflictTable::determine_conflict_table() (RD, 07/12/18)
   */
  std::set<SymbolWithPath> wait_events_; // (RD, 07/12/18)

  /**
   * \brief List of all gathered events this segment is notifying
   *        TODO: this should be computed separately, not as side-effect of
   *        risc::sg::EventConflictTable::determine_conflict_table() (RD, 07/12/18)
   */
  std::set<SymbolWithPath> notify_events_; // (RD, 07/12/18)

  /**
   * \brief This segment calls a function with is marked as conflict
   */
  bool calls_conflict_function_;

  /**
   * \fn    get_min_waiting_time
   * \brief This function determines the minimal waiting time for this segment
   */
  Time get_min_waiting_time();

  /**
   * \brief If this segment has a pointer read
   */
  bool deref_pointer_read_;
  /**
   * \brief If this segment has a pointer write
   */
  bool deref_pointer_write_;

  /**
   * \brief If this segment has an sc_stop, then in conflict with every segment
   */
  bool contains_sc_stop_;

  /**
   * \brief If this segment waits for time
   */
  bool waits_for_time_;

private:
  /**
   * \fn    waits_for_time
   * \brief This functions determines if a function waits for time
   *        The related flag waits_for_time_ will be set during construction
   */
  bool waits_for_time();

  /**
   * \fn    generic_payload_analysis_helper(SgVariableSymbol* var_sym, bool is_write, PortCallPath pcp_of_var)
   * \brief This functions is an internal helper function
   *        It adds the mapped gp data to the read/write_variables
   */
  void generic_payload_analysis_helper(SgVariableSymbol* var_sym, bool is_write, PortCallPath pcp_of_var);

  Time min_waiting_time_;
};

} // end of namespace sg

} // end of namespace risc

#endif /* SEGMENT_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
